from database_generator import DataBaseRandomGenerator
from table import Table
from attribute import Attribute
from attribute_type import FirstNameAttr, LastNameAttr, DigitIDAttr, DateAttr, CityAttr

BookTable = Table(
    {
        "id": Attribute(DigitIDAttr),
        "title": Attribute(CityAttr),
        "author_name": Attribute(FirstNameAttr),
    }
)

Reader = Table(
    {
        "id": Attribute(DigitIDAttr),
        "book_id": Attribute(DigitIDAttr),
        "return_date": Attribute(DateAttr),
    }
)

Person = Table(
    {
        "id": Attribute(DigitIDAttr),
        "Name": Attribute(FirstNameAttr),
        "Surname": Attribute(LastNameAttr),
        "Birthdate": Attribute(DateAttr),
    }
)


BookTable.add_relation("id", "Reader", "book_id", (0, 2))
Reader.add_relation("book_id", "BookTable", "id", (1, 1))

Reader.add_relation("id", "Person", "id", (1, 1))
Person.add_relation("id", "Reader", "id", (0, 1))


db_gen = DataBaseRandomGenerator()

db_gen.add_table("BookTable", BookTable)
db_gen.add_table("Reader", Reader)
db_gen.add_table("Person", Person)

db_gen.random_generation(min_rows=100)

db_gen.tables["BookTable"].to_csv("book_table.csv")
db_gen.tables["Reader"].to_csvdb_gen("reader.csv")
db_gen.tables["Person"].to_csv("person.csv")



import re
from random import random


class Attribute:
    def __init__(self, attr_type, translation_required=False, null_proba=0, **kwargs_rd_gen):

        self.attr_type = attr_type
        self.kwargs_rd_gen = kwargs_rd_gen
        self.null_proba = null_proba
        self.translation_required = translation_required

        self._relation_types = [0, 1, 2]

    def get_random_value(self, fake_factory, translator, locale):

        if random() > self.null_proba:
            return self.attr_type.get_random(
                fake_factory,
                translator,
                locale,
                translation_required=self.translation_required,
                **self.kwargs_rd_gen
            )
        else:
            return None



import csv
import os
import re
import multiprocessing as mp
import pandas as pd
import requests
from bs4 import BeautifulSoup
from tqdm import tqdm

from warnings import simplefilter

simplefilter("ignore")

MIMIC_FILES_DEFAULT_URL = "https://alpha.physionet.org/files/mimiciii-demo/1.4/"


class MimicPreprocessing:
    def __init__(self, dest_folder="data"):
        self.dest_folder = dest_folder
        if not os.path.isdir(dest_folder):
            os.mkdir(dest_folder)
            os.mkdir(f"{dest_folder}/csv")
            os.mkdir(f"{dest_folder}/csv_unique_values")

    def scraper(self, url=MIMIC_FILES_DEFAULT_URL, unused_tables=[]):

        print("Starts scraping mimic...")

        resp = requests.get(url)
        soup = BeautifulSoup(resp.content, "html.parser")
        filename_list = [x.get("href") for x in soup.find_all("a")]
        print(filename_list)
        i = 0
        progress_files = tqdm(filename_list)
        for filename in progress_files:
            progress_files.set_description(f"Fetching {filename}...")
            if not filename.split(".csv")[0].lower() in unused_tables:

                if filename.endswith(".csv"):

                    resp = requests.get(url + filename)
                    with open(f"{self.dest_folder}/csv/{filename.lower()}", "w+") as f:
                        f.write(resp.text)

                i += 1

        print("\nScrapping terminated !\n")

    def infer_attributes(self):
        print("Starts inferring attributes from mimic...")

        attribute_types_all_tables = {}
        possible_values_all_tables = {}
        progress_files = tqdm(os.listdir(f"{self.dest_folder}/csv_unique_values"))
        for filename in progress_files:
            progress_files.set_description(f"Infering attributes from {filename}...")
            if filename.endswith(".csv"):

                attribute_types = {}
                possible_values = {}
                df = pd.read_csv(f"{self.dest_folder}/csv_unique_values/{filename}")
                df.columns = df.columns.str.lower()
                for attr in df.columns:
                    if (
                        False
                    ):  ############################################ TO FIX WHEN ATTRIBUTE TYPE DETECTION HAS BEEN SUFFICIENTLY IMPROVED (bad guessed type may lead to troubles with relations)
                        if re.match(r"^\d+$", attr_values[i]):
                            float_type = False
                            ev_digit_id_type = False
                            relative_float_type = False
                            length = len(attr_values[i])
                            is_field = False
                            for j in range(i + 1, len(attr_values)):
                                if attr_values[j] != "":
                                    if not re.match(r"^\d+$", attr_values[j]):
                                        if re.match(r"^\d+\.\d+$", attr_values[j]):
                                            float_type = True
                                        elif re.match(r"^-\d+\.?\d*$", attr_values[j]):
                                            relative_float_type = True
                                        elif re.match(r"^[EV]?\d+$", attr_values[j]):
                                            ev_digit_id_type = True
                                        else:
                                            attribute_types[attr] = f"field-{attr}"
                                            possible_values[f"{attr}"] = attr_values
                                            is_field = True
                                            break
                                    else:
                                        length = max(length, len(attr_values[j]))
                            if not is_field:
                                if ev_digit_id_type:
                                    attribute_types[attr] = f"ev_digit_id-"
                                elif relative_float_type:
                                    min_value = min([float(x) for x in attr_values if float(x) > 0])
                                    max_value = max([float(x) for x in attr_values if float(x) > 0])
                                    attribute_types[
                                        attr
                                    ] = f"relative_float-{min_value}-{max_value}"
                                elif float_type:
                                    min_value = min([float(x) for x in attr_values])
                                    max_value = max([float(x) for x in attr_values])
                                    attribute_types[attr] = f"float-{min_value}-{max_value}"
                                else:
                                    attribute_types[attr] = f"digit_id-{length}"
                        elif (
                            re.match(r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}$", attr_values[i],)
                            or attr_values[i] == "nan"
                        ):
                            is_field = False
                            for j in range(i + 1, len(attr_values)):
                                if attr_values[j] != "":
                                    if (
                                        not re.match(
                                            r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}$",
                                            attr_values[j],
                                        )
                                        or attr_values[i] == "nan"
                                    ):
                                        attribute_types[attr] = f"field-{attr}"
                                        possible_values[f"{attr}"] = attr_values
                                        is_field = True
                                        break
                            if not is_field:
                                attribute_types[attr] = f"date-"
                        elif re.match(r"^\d+\.\d+$", attr_values[i]) or attr_values[i] == "nan":
                            relative_float_type = False
                            is_field = False
                            for j in range(i + 1, len(attr_values)):
                                if attr_values[j] != "":
                                    if (
                                        not re.match(r"^\d+\.?\d*$", attr_values[j])
                                        or attr_values[j] == "nan"
                                    ):
                                        if re.match(r"^-\d+\.?\d*$", attr_values[j]):
                                            relative_float_type = True
                                        else:
                                            attribute_types[attr] = f"field-{attr}"
                                            possible_values[f"{attr}"] = attr_values
                                            is_field = True
                                            break
                            if not is_field:
                                if relative_float_type:
                                    min_value = min([float(x) for x in attr_values if float(x) > 0])
                                    max_value = max([float(x) for x in attr_values if float(x) > 0])
                                    attribute_types[
                                        attr
                                    ] = f"relative_float-{min_value}-{max_value}"
                                else:
                                    min_value = min([float(x) for x in attr_values])
                                    max_value = max([float(x) for x in attr_values])
                                    attribute_types[attr] = f"float-{min_value}-{max_value}"
                        elif re.match(r"^-\d+\.?\d*$", attr_values[i]) or attr_values[i] == "nan":
                            is_field = False
                            for j in range(i + 1, len(attr_values)):
                                if attr_values[j] != "":
                                    if (
                                        not re.match(r"^-\d+\.?\d*$", attr_values[j])
                                        or attr_values[j] == "nan"
                                    ):
                                        attribute_types[attr] = f"field-{attr}"
                                        possible_values[
                                            f"{filename.split('.')[0]}-{attr}"
                                        ] = attr_values
                                        is_field = True
                                        break
                            if not is_field:
                                min_value = min(
                                    [abs(float(x)) for x in attr_values if float(x) > 0]
                                )
                                max_value = max([float(x) for x in attr_values if float(x) > 0])
                                attribute_types[attr] = f"relative_float-{min_value}-{max_value}"
                        elif re.match(r"^\d{5}-\d{5}$", attr_values[i]):
                            is_field = False
                            for j in range(i + 1, len(attr_values)):
                                if attr_values[j] != "":
                                    if not re.match(r"^\d{5}-\d{5}$", attr_values[j]):
                                        attribute_types[attr] = f"field-{attr}"
                                        possible_values[f"{attr}"] = attr_values
                                        is_field = True
                                        break
                            if not is_field:
                                attribute_types[attr] = f"double_digit_id-5"
                        elif re.match(r"^\d+-\d$", attr_values[i]) or attr_values[i] == "nan":
                            is_field = False
                            for j in range(i + 1, len(attr_values)):
                                if attr_values[j] != "":
                                    if (
                                        not re.match(r"^\d+-\d$", attr_values[j])
                                        or attr_values[j] == "nan"
                                    ):
                                        attribute_types[attr] = f"field-{attr}"
                                        possible_values[f"{attr}"] = attr_values
                                        is_field = True
                                        break
                            if not is_field:
                                attribute_types[attr] = f"composed_digit_id-"
                        elif re.match(r"^[EV]?\d+$", attr_values[i]):
                            is_field = False
                            for j in range(i + 1, len(attr_values)):
                                if attr_values[j] != "":
                                    if not re.match(r"^[EV]?\d+$", attr_values[j]):
                                        attribute_types[attr] = f"field-{attr}"
                                        possible_values[f"{attr}"] = attr_values
                                        is_field = True
                                        break
                            if not is_field:
                                attribute_types[attr] = f"ev_digit_id-"
                    else:
                        attribute_types[attr] = f"field-{attr}"
                        possible_values[f"{attr}"] = (
                            df[attr].dropna().reset_index(drop=True).map(str)
                        )

                attribute_types_all_tables[filename.split(".")[0].lower()] = attribute_types
                possible_values_all_tables[filename.split(".")[0].lower()] = possible_values

        print("\nInferring terminated !\n")
        return attribute_types_all_tables, possible_values_all_tables

    def compute_unique_values(self):
        print("Computing unique attribute values from mimic...")

        with mp.Pool(
            processes=mp.cpu_count()
        ) as pool:  # start as many worker processes as there are CPUs
            workers = pool.starmap_async(
                compute_csv_unique_values,
                [
                    (self.dest_folder, filename)
                    for filename in os.listdir(f"{self.dest_folder}/csv")
                ],
                callback=lambda result: print(f"computed unique values from {result}..."),
            )
            workers.wait()


def compute_csv_unique_values(dest_folder, filename, max_rows=10000000):
    print("processing", filename)
    if filename.endswith(".csv"):

        df = pd.read_csv(f"{dest_folder}/csv/{filename}", nrows=max_rows).applymap(str)
        unique_values = dict()
        for attr in df.columns:
            unique_values[attr] = df[attr].drop_duplicates().reset_index(drop=True)

        with open(f"{dest_folder}/csv_unique_values/{filename}", "w+") as f:
            writer = csv.writer(f)
            writer.writerow(unique_values.keys())
            max_len = max(len(unique_values[x]) for x in df.columns)
            i = 0
            while i < max_len:
                row = []
                for attr in df.columns:
                    values = unique_values[attr]
                    if i < len(values):
                        row.append(values[i])
                    else:
                        row.append("")
                writer.writerow(row)
                i += 1
    return filename



from .attribute import Attribute
import pandas as pd


class Table:
    def __init__(self, attributes_dict):
        self.attributes = attributes_dict
        self.data = pd.DataFrame(columns=self.attributes.keys())

    def add_row(self, row_values):
        self.data = self.data.append(row_values, ignore_index=True)

    def get_random_row(self, fake_factory, translator, locale, fixed_values={}):
        row_values = {}
        for name, value in fixed_values.items():
            row_values[name] = value
        for name, attr in self.attributes.items():
            if not name in row_values.keys():
                row_values[name] = attr.get_random_value(fake_factory, translator, locale)
        return row_values

    def add_random_row(self, fake_factory, translator, locale, fixed_values={}):
        self.add_row(
            self.get_random_row(fake_factory, translator, locale, fixed_values=fixed_values)
        )

    def to_csv(self, output_path):
        self.data.to_csv(output_path, index=False)

    def to_sql(self, table_name, engine):
        self.data.to_sql(table_name, engine, index=False)

    def length(self):

        return len(self.data)



import os
from fpdf import FPDF
from faker import Factory
from random import randint, choice
import pandas as pd
from unidecode import unidecode
from path import Path
from warnings import simplefilter
from datetime import datetime, timedelta

simplefilter("ignore")

MEDIC = [
    "DOLIPRANE",
    "DAFALGAN",
    "IMODIUM",
    "SPASFON",
    "TAHOR",
    "SPEDIFEN",
    "VOLTARENE",
    "ELUDRIL",
    "IXPRIM",
    "PARACETAMOL BIOGARAN",
    "FORLAX",
    "MAGNE B6",
    "HELICIDINE",
    "CLAMOXYL",
    "PIASCLEDINE",
    "LAMALINE",
    "GAVISCON",
    "DAFLON",
    "ANTARENE",
    "RHINOFLUIMUCIL",
    "PLAVIX",
    "MOPRAL",
    "SUBUTEX",
    "AERIUS",
    "ORELOX",
    "INEXIUM",
    "METEOSPASMYL",
    "AUGMENTIN",
    "TOPLEXIL",
    "PIVALONE",
    "VASTAREL",
    "ADVIL",
    "EUPANTOL",
    "DEXERYL",
    "RENUTRYL 500",
    "XANAX",
    "EMLAPATCH",
    "LASILIX",
    "ENDOTELON",
    "DEROXAT",
    "TEMESTA",
    "EFFEXOR",
    "PARACETAMOL SANDOZ",
    "VENTOLINE",
    "SOLUPRED",
    "DEXTROPROPOXYPHENE PARAC BIOG",
    "PNEUMOREL",
    "INIPOMP",
    "PREVISCAN",
    "ASPEGIC",
    "GINKOR",
    "CRESTOR",
    "MEDIATOR",
    "SERESTA",
    "MOTILIUM",
    "PARACETAMOL MERCK",
    "CELESTENE",
    "AMLOR",
    "DIAMICRON",
    "TANAKAN",
    "ATARAX",
    "DERINOX",
    "XYZALL",
    "DEXTROPROPOXYPHENE PARAC SAND",
    "SERETIDE",
    "COVERSYL",
    "PROPOFAN",
    "HEXAQUINE",
    "APROVEL",
    "PARIET",
    "ZALDIAR",
    "DIPROSONE",
    "PARACETAMOL TEVA",
    "BETADINE",
    "LYSANXIA",
    "ALODONT",
    "LEXOMIL",
    "DACRYOSERUM",
    "FUCIDINE",
    "STILNOX",
    "KETUM",
    "STABLON",
    "ART",
    "BIOCALYPTOL",
    "THIOVALONE",
    "DEBRIDAT",
    "PYOSTACINE",
    "TIORFAN",
    "SPECIAFOLDINE",
    "OGAST",
    "RIVOTRIL",
    "TOPALGIC",
    "NASONEX",
]

POSO_TYPE = ["cp", "injections", "comprimés"]

POSO_FREQ = [
    "le matin",
    "le soir",
    "matin, midi et soir",
    "avant les repas",
    "après les repas",
    "par jour",
]

POSO_DURATION = [f"pendant {n} jours" for n in {3, 5, 7, 10, 15, 30}] + ["si douleur"]

MONTHS = [
    "janvier",
    "février",
    "mars",
    "avril",
    "mai",
    "juin",
    "août",
    "septembre",
    "octobre",
    "novembre",
    "décembre",
]

HOSP = [["ischémie de main droite","L intervention a consisté en une ligature de l artère radiale distale" ],
             ["urétéroscopie","Simples, Ablation de la sonde vésicale à J1. Reprise mictionnelle satisfaisante autorisant la sortie"],
             ["superficialisation","L intervention a consisté en une superficialisation de veine radiale droite."],
             ["création d un abord vasculaire","L intervention a consisté en la création d une fistule radio céphalique gauche"],
             ["PUC GENOU GAUCHE", "PUC Genou Gauche sous AG. Suites simples"],
             ["fermeture exérèse d un abord vasculaire", "L intervention a consisté en une fermeture exérèse de l abord vasculaire"],
             ["Hydrocèle scrotale gauche", "Augmentation du volume de la bourse gauche ,  confirmé en échographie"],
             ["coronarographie de contrôle","La coronarographie réalisée par voie radiale droite retrouve une bonne perméabilité des stents IVA et circonflexe"],
             ["coronarographie diagnostique", "La coronarographie retrouve une atteinte coronaire tri tronculaire"],
             ["tentative de désobstruction IVA","La procédure a consisté en une tentative de recanalisation de l IVA avec échec"],
             ["pontage aorto coronaire", "L intervention consiste en un double pontage IVA"],
             ["artériographie des membres inférieurs", "L artériographie retrouve une arthrite distale sévère"],
             ["faux anévrisme de l aorte ascendante", "Le suivi scanographique a montré l apparition d un aspect d un double faux anévrisme de l aorte initiale"]]

ANT = ["fumeur",
       "hypertension artérielle",
       "hypercholestérolémie",
       "diabète",
       "diabète de type 2",
       "obésité",
       "Tabac/PA",
        "Insuffisance rénale chronique",
        "Surpoids",
        "Pas d’allergie"]

FONTS_FOLDER_PATH = Path(os.path.dirname(__file__)) / "utils/fonts/"

SIGNATURE_FONTS = [
    FONTS_FOLDER_PATH / name
    for name in os.listdir(FONTS_FOLDER_PATH)
    if name.endswith(".ttf")
]

MAIN_FONTS = ["Arial", "Courier", "Helvetica", "Times"]


def add_block(pdf, txt, align="L"):
    """Adds paragraph to the pdf object

    Arguments:
        pdf {fpdf.fpdf.FPDF} -- current pdf object
        txt {str} -- paragraph to add

    Keyword Arguments:
        align {str} -- text alignment : the possible values are "L", "C" and "R" (default: {"L"})
    """

    for row in txt.split("\n"):
        pdf.cell(200, 5, txt=row, ln=1, align=align)


def signature(name):
    """Creates a fake signature

    Arguments:
        name {str} -- person name

    Returns:
        str -- signature
    """

    name_list = unidecode(name).split(" ")
    sign = ""
    for w in name_list[:-1]:
        if randint(0, 1):
            if randint(0, 1):
                sign += w[0]
            else:
                sign += w
            if randint(0, 1):
                sign += " "
    sign += name_list[-1]
    return sign


def prescription_generator(output_path, locale="fr_FR", **kwargs):
    """Generates a fake medical prescription

    Arguments:
        output_path {Path} -- path the ouput pdf file
        patient_name {str} -- patient name to use
        year {int} -- year of the prescription
    """

    fake = Factory.create(locale)
    pdf = FPDF(format="A4")
    pdf.add_page()
    font = choice(MAIN_FONTS)
    pdf.set_font(font, size=11)

    # Arguments processing

    try:
        nb_medics = len(kwargs["medication"])
    except KeyError:
        nb_medics = randint(1, 5)

    field_group_values = {
        "doc_attributes": {"city": fake.city(), "date": fake.date(pattern="%d/%m/%Y"),},
        "medical_centre": {
            "city": fake.city(),
            "street_address": fake.street_address(),
            "postal_code": randint(1000, 10000) * 10,
            "phone_number": fake.phone_number(),
        },
        "doctor": {
            "first_name": fake.first_name(),
            "last_name": fake.last_name(),
            "street_address": fake.street_address(),
            "postal_code": randint(1000, 10000) * 10,
            "city": fake.city(),
            "phone_number": fake.phone_number(),
        },
        "patient": {
            "first_name": fake.first_name(),
            "last_name": fake.last_name(),
            "age": randint(10, 100),
            "weight": randint(30, 110),
        },
        "medication": [],
    }

    for _ in range(nb_medics):
        field_group_values["medication"].append(
            {
                "name": choice(MEDIC),
                "dose": 5 * randint(1, 10),
                "poso_qty": randint(1, 3),
                "poso_type": choice(POSO_TYPE),
                "poso_freq": choice(POSO_FREQ),
                "poso_duration": choice(POSO_DURATION),
            }
        )

    for group_name in {"doc_attributes", "medical_centre", "doctor", "patient"}:
        for field_name in field_group_values[group_name]:
            try:
                field_group_values[group_name][field_name] = kwargs[group_name][
                    field_name
                ]
            except KeyError:
                pass

    if "medication" in kwargs.keys():
        for i in range(nb_medics):
            for field_name in field_group_values["medication"][i]:
                try:
                    field_group_values["medication"][i][field_name] = kwargs[
                        "medication"
                    ][i][field_name]
                except KeyError:
                    pass

    for group_name, group_values in kwargs.items():
        if not group_name in field_group_values:
            print(f"Attributes group '{group_name}' is not recognized.")
        elif group_name in {"doc_attributes", "medical_centre", "doctor", "patient"}:
            for field_name in group_values.keys():
                if not field_name in field_group_values[group_name]:
                    print(
                        f"Attribute '{field_name}' from group '{group_name}' is not recognized."
                    )
        else:
            assert group_name == "medication", f"Unexpected argument."
            assert (
                type(kwargs["medication"]) == list
            ), f"Wrong argument type : 'medication' argument must be a list of dictionnaries."
            for i in range(nb_medics):
                for field_name in kwargs["medication"][i].keys():
                    if not field_name in field_group_values["medication"][i].keys():
                        print(
                            f"Attribute '{field_name}' from medication #{i} is not recognized."
                        )

    # En-tête

    txt = f"""À {field_group_values["doc_attributes"]["city"]}, le {field_group_values["doc_attributes"]["date"]}     
    
    
    
    """
    add_block(pdf, txt, align="R")

    # Centre médical

    txt = f"""
Centre Médical de {field_group_values["medical_centre"]["city"]}
{field_group_values["medical_centre"]["street_address"]}
{field_group_values["medical_centre"]["postal_code"]} {field_group_values["medical_centre"]["city"]}
{field_group_values["medical_centre"]["phone_number"]}


"""
    pdf.set_font(font, "B", size=11)
    add_block(pdf, txt, align="C")
    pdf.set_font(font, size=11)

    # Docteur

    txt = f"""
        Docteur {field_group_values["doctor"]["last_name"]}
        {field_group_values["doctor"]["street_address"]}
        {field_group_values["doctor"]["postal_code"]} {field_group_values["doctor"]["city"]}
        Tél. : {field_group_values["doctor"]["phone_number"]}

"""
    add_block(pdf, txt)

    # Patient

    txt = f"""
        {field_group_values["patient"]["first_name"]} {field_group_values["patient"]["last_name"]}
        {field_group_values["patient"]["age"]} ans, {field_group_values["patient"]["weight"]} kg

    """

    add_block(pdf, txt)

    # Prescription

    for i in range(nb_medics):

        txt = f"""
        {field_group_values["medication"][i]["name"]} {field_group_values["medication"][i]["dose"]}"""

        pdf.set_font(font, "B", size=11)
        add_block(pdf, txt)

        txt = f"""          {field_group_values["medication"][i]["poso_qty"]} {field_group_values["medication"][i]["poso_type"]} {field_group_values["medication"][i]["poso_freq"]} {field_group_values["medication"][i]["poso_duration"]}
"""

        pdf.set_font(font, size=11)
        add_block(pdf, txt)

    # Signature

    pdf.add_font("Signature", "", choice(SIGNATURE_FONTS), uni=True)
    font_size = randint(20, 40)
    pdf.set_font("Signature", size=font_size)

    txt = f"""


{signature(field_group_values["doctor"]['last_name'])}          {int((40 - font_size)/20) * randint(0, 20) * ' '}
"""
    add_block(pdf, txt, align="R")

    pdf.output(output_path)

def cr_generator(output_path, locale="fr_FR", **kwargs):
    """Generates a fake medical report

    Arguments:
        output_path {Path} -- path the ouput pdf file
        patient_name {str} -- patient name to use
        year {int} -- year of the prescription
    """
    fake = Factory.create(locale)
    pdf = FPDF(format="A4")
    pdf.add_page()
    font = choice(MAIN_FONTS)
    pdf.set_font(font, size=10)

    #Arguments processing

    try:
        nb_medics = len(kwargs["medication"])
    except KeyError:
        nb_medics = randint(1, 5)

    try:
        nb_ant = len(kwargs["antecedent"])
    except KeyError:
        nb_ant = randint(1, 3)

    choice_hospitalization = randint(0, len(HOSP)-1)
    last_date = fake.date_between()
    duration = randint(0,16)
    first_date = last_date - timedelta(days=duration)

    field_group_values = {
        "doc_attributes": {"city": fake.city(), "date": fake.date(pattern="%d/%m/%Y"),},
        "medical_centre": {
            "city": fake.city(),
            "street_address": fake.street_address(),
            "postal_code": randint(1000, 10000) * 10,
            "phone_number": fake.phone_number(),
        },
        "doctor": {
            "first_name": fake.first_name(),
            "last_name": fake.last_name(),
            "street_address": fake.street_address(),
            "postal_code": randint(1000, 10000) * 10,
            "city": fake.city(),
            "phone_number": fake.phone_number(),
        },
        "patient": {
            "first_name": fake.first_name(),
            "last_name": fake.last_name(),
            "age": randint(10, 100),
            "weight": randint(30, 110),
        },
        "hospitalization": {
            "reason": HOSP[choice_hospitalization][0],
            "hospitalization": HOSP[choice_hospitalization][1]
        },
        "date": {
            "last_date": last_date.strftime("%d/%m/%Y"),
            "first_date": first_date.strftime("%d/%m/%Y"),
            "birth_date": fake.date_between(start_date='-80y', end_date=first_date).strftime("%d/%m/%Y")
        },
        "medication": [],
        "antecedent": []
    }

    for _ in range(nb_medics):
        field_group_values["medication"].append(
            {
                "name": choice(MEDIC),
                "dose": 5 * randint(1, 10),
                "poso_qty": randint(1, 3),
                "poso_type": choice(POSO_TYPE),
                "poso_freq": choice(POSO_FREQ),
                "poso_duration": choice(POSO_DURATION),
            }
        )

    for _ in range(nb_ant):
        field_group_values["antecedent"].append(choice(ANT))

    for group_name in {'doc_attributes', 'medical_centre', 'doctor', 'patient', "hospitalization", "date"}:
        for field_name in field_group_values[group_name]:
            try:
                field_group_values[group_name][field_name] = kwargs[group_name][
                    field_name
                ]

            except KeyError:
                pass

    if "medication" in kwargs.keys():
        for i in range(nb_medics):
            for field_name in field_group_values["medication"][i]:
                try:
                    field_group_values["medication"][i][field_name] = kwargs[
                        "medication"
                    ][i][field_name]
                except KeyError:
                    pass
    if "antecedent" in kwargs.keys():
        for i in range(nb_ant):
            try:
                field_group_values["antecedent"][i] = kwargs["antecedent"][i]
            except:
                pass

    for group_name, group_values in kwargs.items():
        if not group_name in field_group_values:
            print(f"Attributes group '{group_name}' is not recognized.")
        elif group_name in {'doc_attributes', 'medical_centre', 'doctor', 'patient', 'date', 'hospitalization'}:
            for field_name in group_values.keys():
                if not field_name in field_group_values[group_name]:
                    print(
                        f"Attribute '{field_name}'' from group '{group_name}' is not recognized."
                    )
        elif group_name == 'antecedent':
            pass
        else:
            assert group_name == 'medication', f"Unexpected argument."
            assert type(kwargs[
                            "medication"]) == list, f"Wrong argument type : 'medication' argument must be a list of dictionnaries."
            for i in range(nb_medics):
                for field_name in kwargs["medication"][i].keys():
                    if not field_name in field_group_values['medication'][i].keys():
                        print(
                            f"Attribute '{field_name}' from medication #{i} is not recognized."
                        )

    # Centre médical

    txt = f"""
Centre Médical de {field_group_values["medical_centre"]["city"]}
{field_group_values["medical_centre"]["street_address"]}
{field_group_values["medical_centre"]["postal_code"]} {field_group_values["medical_centre"]["city"]}
{field_group_values["medical_centre"]["phone_number"]}


"""
    pdf.set_font(font, "B", size=11)
    add_block(pdf, txt, align="L")
    pdf.set_font(font, size=10)

    # Docteur

    doctor_name = fake.name()
    postal_code = randint(1000, 10000) * 10

    txt = f"""
Docteur {field_group_values["doctor"]["last_name"]}
{field_group_values["doctor"]["street_address"]}
{field_group_values["doctor"]["postal_code"]} {field_group_values["doctor"]["city"]}
Tél. : {field_group_values["doctor"]["phone_number"]}

"""
    add_block(pdf, txt)

    # Date
    txt = f"""À {field_group_values["medical_centre"]["city"]}, le {field_group_values["date"]["last_date"]}     

        """
    add_block(pdf, txt, align="R")

    #Objet
    txt = f"""Objet : Compte-rendu d'hospitalisation de {field_group_values["patient"]["first_name"]} {field_group_values["patient"]["last_name"]} du {field_group_values["date"]["first_date"]} au {field_group_values["date"]["last_date"]}

{field_group_values["patient"]["first_name"]} {field_group_values["patient"]["last_name"]} né le {field_group_values["date"]["birth_date"]} a été hospitalisé le {field_group_values["date"]["first_date"]} pour {field_group_values["hospitalization"]["reason"]}

"""
    add_block(pdf, txt, align="L")

    #Antécédents
    pdf.set_font(font, "B", size=10)
    txt = """
Antécédents
    """
    add_block(pdf, txt, align="L")
    pdf.set_font(font, size=10)
    for i in range(nb_ant):
        txt = f"""      {field_group_values["antecedent"][i]} """
        add_block(pdf, txt)

    #Hospitalisation
    pdf.set_font(font, "B", size=10)
    txt = """
Hospitalisation
    """
    add_block(pdf, txt, align="L")
    pdf.set_font(font, size=10)
    txt = f"""      {field_group_values["hospitalization"]["hospitalization"]}"""
    add_block(pdf, txt, align="L")

    #Traitement de sortie
    pdf.set_font(font, "B", size=10)
    txt = """
Traitement de sortie
    """
    add_block(pdf, txt, align="L")
    pdf.set_font(font, size=10)
    for i in range(nb_medics):
        txt = f"""
        {field_group_values["medication"][i]["name"]} {field_group_values["medication"][i]["dose"]} mg, {field_group_values["medication"][i]["poso_qty"]} {field_group_values["medication"][i]["poso_type"]} {field_group_values["medication"][i]["poso_freq"]} {field_group_values["medication"][i]["poso_duration"]}"""
    add_block(pdf, txt, align="L")
    # Signature

    pdf.add_font("Signature", "", choice(SIGNATURE_FONTS), uni=True)
    font_size = randint(20, 40)
    pdf.set_font("Signature", size=font_size)

    txt = f"""
    {signature(field_group_values["doctor"]['last_name'])}          {int((40 - font_size) / 20) * randint(0, 20) * ' '}
    """
    add_block(pdf, txt, align="R")

    pdf.output(output_path)


def get_id():
    """Generates a random 6-digit id

    Returns:
        str -- random 6-digit id
    """

    return str(randint(1000, 100000)).zfill(6)


def generate(nb, output_path):
    """Generate several fake medical prescriptions

    Arguments:
        nb {int} -- number or fake medical prescriptions to generate
        output_path {str} -- folder to store the fake medical prescriptions in
    """

    output_path = Path(output_path)

    link = pd.DataFrame(
        columns=[
            "patient_id",
            "patient_surname",
            "patient_name",
            "date_folder",
            "prescription_id",
            "cr_id",
        ]
    )

    print("Generating data...")

    for i in range(nb):

        print(f"Loading : {int(100 * i / nb)} %", end="\r")

        patient_id = get_id()
        while not link[lambda row: row["patient_id"] == patient_id].empty:
            patient_id = get_id()

        prescription_id = get_id()
        while not link[lambda row: row["prescription_id"] == prescription_id].empty:
            prescription_id = get_id()

        cr_id = get_id()
        while not link[lambda row: row["prescription_id"] == prescription_id].empty:
            cr_id = get_id()

        patient_name = fake.first_name()
        patient_surname = fake.last_name()
        year = str(randint(2010, 2019))

        link = link.append(
            {
                "patient_id": patient_id,
                "patient_surname": patient_surname,
                "patient_name": patient_name,
                "date_folder": year,
                "prescription_id": prescription_id,
                "cr_id": cr_id,
            },
            ignore_index=True,
        )

        if not os.path.isdir(output_path / year):
            os.mkdir(output_path / year)

        prescription_generator(
            f"{output_path}/{year}/prescription_{prescription_id}.pdf",
            patient={"first_name": patient_name, "last_name": patient_surname},
            doc_attributes={"date": f"01/01/{year}"},
        )
        cr_generator(
            f"{output_path}/{year}/CR_{cr_id}.pdf",
            patient = {"first_name":patient_name, "last_name": patient_surname},
            doc_attributes={'date': f'01/01/{year}'},
        )

    link.to_csv(output_path / "index.csv")

    print("Generation terminated !")


import os
import csv
import argparse
from path import Path
from unidecode import unidecode
import json
import sys
import re

from .extract import extract


def search(query, input_path, output_path, scale="row", update_cache=True):
    """Searchs a query in extracted files the given input folder and its subfolders and stores the results at the given output folder.

    Arguments:
        query {str} -- query : words must be separated with spaces, for mandatory words use +, for forbidden words use -, for complex expressions use ""
        input_path {str} -- folder to search in
        output_path {str} -- folder to store the results in

    Keyword Arguments:
        scale {str} -- scale of the research : the possible values are "row" and "doc" (default: {"row"})
        update_cache {bol} -- `True` to update the cached files (default: {False})
    
    Returns:
        None
    """

    if update_cache:
        extract(input_path)

    input_path = Path(input_path)
    output_path = Path(output_path)

    assert scale in {
        "row",
        "doc",
    }, f"Wrong scale value. It must be row or doc, not {scale}."

    query = query.split(" ")

    clean_query = preprocess(query)
    result, count_dict = search_folder(clean_query, input_path, scale, first_call=True)

    output_path = output_path / normalize("__".join(query)).replace("/", "").replace(
        " ", "_"
    )

    if not os.path.isdir(output_path):
        os.mkdir(output_path)

    print(f"Found {count_dict['nb']} results.")

    with open(output_path / "folders.json", "w+",) as f:
        json.dump(count_dict, f, indent=4)

    with open(output_path / "results.csv", "w+",) as f:
        writer = csv.writer(f)
        for row in result:
            if scale == "row":
                writer.writerow(row)
            else:
                writer.writerow(row[0])


def search_folder(query, path, scale, first_call=False):
    """Recursively searchs a query in a given folder at a given scale

    Arguments:
        query {list[list[str]]} -- preprocessed query to search
        path {Path} -- folder to search in
        scale {row} -- scale of the research : the possible values are "row" and "doc" (default: {"row"})

    Keyword Arguments:
        first_call {bool} -- for monitoring purpose : must be True if the current call is not recursive (default: {False})

    Returns:
        dict -- the key 'nb' refers to the total number of occurrences recursively found in the given folder, the key 'result' is the list of the results
    """

    path = Path(path)

    if first_call:
        print("New research")
        count_folder = 0

    result = [["path", "row", "context"]]
    count_dict = {}
    count = 0
    targets = os.listdir(path)

    for f in targets:

        current_path = path / f

        if os.path.isdir(current_path):

            if first_call:
                print(f"Loading : {int(100 * count_folder / len(targets))} %", end="\r")
                count_folder += 1

            local_result, local_dict = search_folder(query, current_path, scale)
            result += local_result[1:]
            count += local_dict["nb"]

            if f in count_dict.keys():
                count_dict[f]["nb"] += local_dict["nb"]
                count_dict[f]["path"].update(local_dict["path"])
            else:
                count_dict[f] = local_dict

        else:
            if re.match(r"\.\w{2,4}\-{5}.+", f.lower()):
                local_result = search_file(query, current_path, scale)[1:]
                result += local_result
                count += len(local_result)

    if len(count_dict) > 0:
        return result, {"nb": count, "path": count_dict}
    else:
        return result, {"nb": count}


def search_file(query, path, scale):
    """Searchs a query in a given file at a given scale

    Arguments:
        query {list[list[str]]} -- preprocessed query to search
        path {Path} -- file to search in
        scale {row} -- scale of the research : the possible values are "row" and "doc" (default: {"row"})

    Returns:
        list -- list of occurrences found. For each occurrence, we store the file path, the row number and the context. In 'doc' scale, the row number and the context are None.
    """

    path = Path(path)
    result = [["path", "row", "context"]]
    with open(path, "r") as f:
        if scale == "row":
            for i, row in enumerate(f):
                if search_str(query, row):
                    file_type, file_name = re.findall(
                        r"\.(\w{2,4})\-{5}(.+\.)txt", path
                    )[0]
                    result.append(
                        [
                            path.rsplit("/", 1)[0] + "/" + file_name + file_type,
                            i,
                            row[:-1],
                        ]
                    )
        elif scale == "doc":
            txt = f.read()
            if search_str(query, txt):
                result.append([str(path), None, None])
    return result


def search_str(query, txt):
    """Searchs a query in a given string

    Arguments:
        query {list[list[str]]} -- preprocessed query to search
        txt {str} -- string to search in

    Returns:
        bool -- True if an occurrence has been found
    """

    for token in query[0]:
        if normalize(txt).find(normalize(token)) < 0:
            return False
    for token in query[1]:
        if normalize(txt).find(normalize(token)) >= 0:
            return False
    for token in query[2]:
        if normalize(txt).find(normalize(token)) >= 0:
            return True
    return False


def normalize(txt):
    """Normalizes text

    Arguments:
        txt {str} -- text to normalize

    Returns:
        str -- normalized text
    """

    assert type(txt) == str
    return unidecode(txt.lower())


def preprocess(args_list):
    """Preprocesses a query

    Arguments:
        args_list {list[str]} -- list of words (for mandatory words use +, for forbidden words use -, for complex expressions use)

    Returns:
        list[lst[str]] -- preprocessed query
    """

    OR = []
    AND = []
    NOT = []

    i = 0
    while i < len(args_list):
        arg = args_list[i]

        if arg[0] not in {"+", "-"}:
            arg = "§" + arg

        if arg[1] in {"'", '"'}:
            part = arg

            while not part[-1] in {'"', "'"}:

                if i >= len(args_list) - 1:
                    raise Exception("missing quote mark in the query.")

                i += 1
                part = args_list[i]
                arg += " " + part

            arg = arg[0] + arg[2:-1]

        if arg[0] == "+":
            AND.append(arg[1:])
        elif arg[0] == "-":
            NOT.append(arg[1:])
        else:
            assert arg[0] == "§"
            OR.append(arg[1:])

        i += 1

    return [AND, NOT, OR]


if __name__ == "__main__":

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "query",
        metavar="Q",
        type=str,
        nargs="+",
        help='your query with several words seperated by spaces. Use + for mandatory words, - for forbidden words and "" for complex expressions.',
    )
    parser.add_argument(
        "--d",
        dest="doc_scale",
        action="store_true",
        default=False,
        help="to search at the doc scale instead of at the row scale",
    )
    parser.add_argument(
        "--input_path",
        dest="input_path",
        action="store",
        help="path to the folder to search in",
    )
    parser.add_argument(
        "--output_path",
        dest="output_path",
        action="store",
        help="path to the folder to put the results in",
    )

    args = parser.parse_args()

    scale = "doc" if args.doc_scale else "row"

    search(" ".join(args.query), args.input_path, args.output_path, scale=scale)



from random import choice, randint, random
from unidecode import unidecode
import pandas as pd
import os
import html


class AttributeType:
    def __init__(self, type_name, random_generator):
        self.type_name = type_name
        self.random_generator = random_generator

    def get_random(self, fake_factory, translator, locale, **kwargs_for_random_generator):
        return self.random_generator(
            fake_factory, translator, locale, **kwargs_for_random_generator
        )


### Random generators


def first_name_random_generator(fake_factory, translator, locale, **kwargs):
    return fake_factory.first_name()


def last_name_random_generator(fake_factory, translator, locale, **kwargs):
    return fake_factory.last_name()


def signature_random_generator(*args, name=None, **kwargs):
    name_list = unidecode(name).split(" ")
    sign = ""
    for w in name_list[:-1]:
        if randint(0, 1):
            if randint(0, 1):
                sign += w[0]
            else:
                sign += w
            if randint(0, 1):
                sign += " "
    return sign + name_list[-1]


def city_random_generator(fake_factory, translator, locale, **kwargs):
    return fake_factory.city()


def postal_code_random_generator(*args, **kwargs):
    return randint(1000, 10000) * 10


def street_address_random_generator(fake_factory, translator, locale, **kwargs):
    return fake_factory.street_address()


def integer_random_generator(*args, min_val=0, max_val=1000, **kwargs):
    return randint(min_val, max_val)


def float_random_generator(*args, min_val=0, max_val=100, **kwargs):
    return min_val + random() * (max_val - min_val)


def relative_float_random_generator(*args, min_val=0, max_val=100, **kwargs):
    if randint(0, 1):
        return randint(-99, -1)
    return float_random_generator(*args, min_val=min_val, max_val=max_val, **kwargs)


def bool_random_generator(*args, proba_true=0.9, **kwargs):
    return random() < proba_true


def digitID_random_generator(*args, length=5, **kwargs):
    min_val = int(10 ** (length - 2))
    max_val = int(10 ** length - 1)
    return str(randint(min_val, max_val)).zfill(length)


def double_digitID_random_generator(*args, length=5, **kwargs):
    min_val = int(10 ** (length - 2))
    max_val = int(10 ** length - 1)
    return (
        str(randint(min_val, max_val)).zfill(length)
        + "-"
        + str(randint(min_val, max_val)).zfill(length)
    )


def EV_digitID_random_generator(*args, digit_length=5, **kwargs):
    return (
        str(choice(["E", "V"] + list(range(10))))
        + 3 * str(randint(0, 9))
        + str(choice([""] + list(range(10))))
    )


def composed_digitID_random_generator(*args, **kwargs):
    return str(randint(100, 99999)) + "-" + randint(0, 9)


def phone_number_random_generator(fake_factory, translator, locale, **kwargs):
    return fake_factory.numerify(choice([f"0{d} ## ## ## ##" for d in range(1, 8)]))


def date_random_generator(
    fake_factory,
    translator,
    locale,
    pattern="%Y-%m-%d %H:%M:%S",
    start_date="-10y",
    end_date="today",
    **kwargs,
):
    return fake_factory.date_between(start_date=start_date, end_date=end_date).strftime(pattern)


def path_random_generator(*args, prefix="", file_type="", **kwargs):
    return f"{prefix}{digitID_random_generator(length=15)}.{file_type}"


def field_random_generator(
    fake_factory, translator, locale, translation_required=False, possible_values=None, **kwargs,
):
    if len(possible_values) == 0:
        return ""

    txt = choice(possible_values)
    language = locale.split("_")[0]
    if language == "en" or translation_required == False:
        return txt
    else:
        if "GOOGLE_APPLICATION_CREDENTIALS" in os.environ.keys():
            try:
                translated_txt = html.unescape(
                    translator.translate(txt, target_language=language)["translatedText"]
                )
                if translated_txt == txt:
                    print(f"A text has not been translated : {txt}")
                    return txt
                return translated_txt
            except:
                print(f"Used initial text because an error occurred while translating : {txt}.")
                return txt
        else:
            raise Exception("Environment variable GOOGLE_APPLICATION_CREDENTIALS not found.")


### Attribute types

FirstNameAttr = AttributeType("first_name", first_name_random_generator)
LastNameAttr = AttributeType("last_name", last_name_random_generator)

CityAttr = AttributeType("city", city_random_generator)
PostalCodeAttr = AttributeType("postal_code", postal_code_random_generator)
StreetAddressAttr = AttributeType("street_address", street_address_random_generator)

IntegerAttr = AttributeType("integer", integer_random_generator)
FloatAttr = AttributeType("float", float_random_generator)
RelativeFloatAttr = AttributeType("relative_float", relative_float_random_generator)
BoolAttr = AttributeType("bool_attr", bool_random_generator)

DigitIDAttr = AttributeType("digit_ID", digitID_random_generator)
DoubleDigitIDAttr = AttributeType("double_digitID", double_digitID_random_generator)
EVDigitIDAttr = AttributeType("EV_digitID", EV_digitID_random_generator)
ComposedDigitIDAttr = AttributeType("composed_digitID", composed_digitID_random_generator)

PhoneNumberAttr = AttributeType("phone_number", phone_number_random_generator)

DateAttr = AttributeType("date", date_random_generator)

PathAttr = AttributeType("path", path_random_generator)

FieldAttr = AttributeType("field", field_random_generator)



import pandas as pd
from sqlalchemy import create_engine
import datetime
from random import randint
from pysin import prescription_generator, cr_generator
from tqdm import tqdm

from pysin.mimic.database_generator import DataBaseRandomGenerator
from pysin.mimic.table import Table
from pysin.mimic.attribute import Attribute
from pysin.mimic.attribute_type import (
    FirstNameAttr,
    LastNameAttr,
    CityAttr,
    PostalCodeAttr,
    StreetAddressAttr,
    IntegerAttr,
    FloatAttr,
    RelativeFloatAttr,
    BoolAttr,
    DigitIDAttr,
    DoubleDigitIDAttr,
    ComposedDigitIDAttr,
    EVDigitIDAttr,
    PhoneNumberAttr,
    DateAttr,
    PathAttr,
    FieldAttr,
)
from pysin.mimic.mimic_preprocessing import MimicPreprocessing, MIMIC_FILES_DEFAULT_URL

import argparse


def scrap():
    parser = argparse.ArgumentParser(description="Scrap MIMIC-III data from alpha.physionet.org")
    parser.add_argument(
        "--dest", required=True, type=str, help="the destination path of the scrapped files",
    )
    parser.add_argument(
        "--url",
        type=str,
        help="the url from where the files are scrapped",
        default=MIMIC_FILES_DEFAULT_URL,
    )
    args = parser.parse_args()

    mimic = MimicPreprocessing(args.dest)
    mimic.scraper(
        url=args.url,
        unused_tables=[
            "callout",
            "cptevents",
            "datetimeevents",
            "d_cpt",
            "inputevents_cv",
            "inputevents_mv",
            "noteevents",
            "outputevents",
            "procedureevents_mv",
        ],
    )


def preprocess():
    parser = argparse.ArgumentParser(
        description="Preprocess MIMIC-III CSV files to compute unique values"
    )
    parser.add_argument(
        "--data", required=True, type=str, help="the destination path of the computed files",
    )
    args = parser.parse_args()

    mimic = MimicPreprocessing(args.data)
    mimic.compute_unique_values()


def generate_database():
    parser = argparse.ArgumentParser(
        description="Populate a postgres database with synthetic medical data based on MIMIC-III"
    )
    parser.add_argument(
        "--data",
        required=True,
        type=str,
        help="the location of the scrapped files and unique values of MIMIC tables",
    )
    parser.add_argument(
        "--db",
        required=True,
        type=str,
        help="the database connexion string (eg: postgresql://user:passsword@host/database)",
    )
    parser.add_argument(
        "--translate",
        action="store_true",
        help="should the mimic database be translated into french",
    )
    parser.add_argument(
        "--rows", required=True, type=int, help="The number of rows to generate",
    )
    parser.add_argument(
        "--locale",
        default="fr_FR",
        type=str,
        help="The locale to use when generating data",
        # see available locales here (https://faker.readthedocs.io/en/master/locales.html)
    )

    args = parser.parse_args()

    mimic = MimicPreprocessing(args.data)

    print("Connecting to postgres...")
    engine = create_engine(args.db)
    connection = engine.connect()
    print("OK")

    attributes_dict, possible_values = mimic.infer_attributes()

    unused_attributes = [  # (table_name, attribute_name)
        ("d_items", "dbsource"),
        ("icustays", "dbsource"),
        ("icustays", "first_wardid"),
        ("icustays", "last_wardid"),
        ("transfers", "dbsource"),
        ("transfers", "prev_wardid"),
        ("transfers", "curr_wardid"),
    ]

    tables_to_add = [
        "hospit_services",
        "documents",
    ]

    attributes_to_add = {  # (table_name, Attribute)
        "patients": [
            ("first_name", Attribute(FirstNameAttr)),
            ("last_name", Attribute(LastNameAttr)),
            ("weight", Attribute(IntegerAttr, min_val=40, max_val=110)),
            ("height", Attribute(IntegerAttr, min_val=130, max_val=210)),
            ("street_address", Attribute(StreetAddressAttr)),
            ("postal_code", Attribute(PostalCodeAttr)),
            ("city", Attribute(CityAttr)),
        ],
        "documents": [
            ("date", Attribute(DateAttr)),
            ("type", Attribute(FieldAttr, possible_values=["prescription", "report"])),
            ("description", Attribute(FieldAttr, possible_values=["NONE"])),  # TODO
            (
                "patient_id",
                Attribute(FieldAttr, possible_values=possible_values["patients"]["subject_id"]),
            ),
            ("statut", Attribute(BoolAttr, proba_true=0.9),),
            ("document", Attribute(PathAttr, prefix="documents/", file_type="pdf"),),
            ("doc_id", Attribute(DigitIDAttr)),
            (
                "hadm_id",
                Attribute(FieldAttr, possible_values=possible_values["admissions"]["hadm_id"]),
            ),
            ("cgid", Attribute(FieldAttr, possible_values=possible_values["caregivers"]["cgid"]),),
        ],
        "hospit_services": [
            ("row_id", Attribute(DigitIDAttr)),
            (
                "careunit",
                Attribute(FieldAttr, possible_values=possible_values["services"]["curr_service"], ),
            ),
        ],
        "caregivers": [
            ("first_name", Attribute(FirstNameAttr)),
            ("last_name", Attribute(LastNameAttr)),
        ],
    }

    attributes_to_translate = [  # (table_name, attr_name)
        ("admissions", "diagnosis"),
        ("drgcodes", "description"),
        ("d_icd_diagnoses", "short_title"),
        ("d_icd_diagnoses", "long_title"),
        ("d_icd_procedures", "short_title"),
        ("d_icd_procedures", "long_title"),
    ]

    tables = {}

    for table in tables_to_add:
        attributes_dict[table] = {}

    for curr_table in attributes_dict:
        curr_attributes = {}
        for attr, attr_class in attributes_dict[curr_table].items():
            if not (curr_table, attr) in unused_attributes:
                try:
                    attr_type, attr_params = attr_class.split("-", 1)
                except:
                    raise Exception(f"Wrong attribute class format : {attr_class}.")
                if attr_type == "digit_id":
                    curr_attributes[attr] = Attribute(DigitIDAttr, length=int(attr_params))
                elif attr_type == "date":
                    curr_attributes[attr] = Attribute(DateAttr)
                elif attr_type == "float":
                    min_val, max_val = attr_params.split("-")
                    curr_attributes[attr] = Attribute(
                        FloatAttr, min_val=float(min_val), max_val=float(max_val)
                    )
                elif attr_type == "relative_float":
                    min_val, max_val = attr_params.split("-")
                    curr_attributes[attr] = Attribute(
                        RelativeFloatAttr, min_val=float(min_val), max_val=float(max_val),
                    )
                elif attr_type == "bool":
                    curr_attributes[attr] = Attribute(BoolAttr, proba_true=float(attr_params), )
                elif attr_type == "double_digit_id":
                    curr_attributes[attr] = Attribute(DoubleDigitIDAttr, length=int(attr_params))
                elif attr_type == "composed_digit_id-":
                    curr_attributes[attr] = Attribute(ComposedDigitIDAttr)
                elif attr_type == "first_name":
                    curr_attributes[attr] = Attribute(FirstNameAttr)
                elif attr_type == "last_name":
                    curr_attributes[attr] = Attribute(LastNameAttr)
                elif attr_type == "city_name":
                    curr_attributes[attr] = Attribute(CityAttr)
                elif attr_type == "postal_code":
                    curr_attributes[attr] = Attribute(PostalCodeAttr)
                elif attr_type == "phone_number":
                    curr_attributes[attr] = Attribute(PhoneNumberAttr)
                elif attr_type == "ev_digit_id":
                    curr_attributes[attr] = Attribute(EVDigitIDAttr, digit_length=attr_params)
                elif attr_type == "field":
                    curr_attributes[attr] = Attribute(
                        FieldAttr, possible_values=possible_values[curr_table][attr_params],
                    )
                else:
                    raise Exception(f"Unknow attribute type : {attr_type}.")
                tables[curr_table.split(".")[0]] = Table(curr_attributes)
                try:
                    attr_type, attr_params = attr_class.split("-", 1)
                except:
                    raise Exception(f"Wrong attribute class format : {attr_class}.")
        if curr_table in attributes_to_add.keys():
            for attr_name, attr in attributes_to_add[curr_table]:
                curr_attributes[attr_name] = attr
        tables[curr_table.split(".")[0]] = Table(curr_attributes)

    print(f"Translation is {'enabled' if args.translate else 'disabled'}")
    for table_name, attr_name in attributes_to_translate:
        tables[table_name].attributes[attr_name].translation_required = args.translate

    relations = {
        "hadm_id": {
            "parents": [("admissions", "hadm_id"), ],
            "children": [
                ("chartevents", "hadm_id", (1, 2)),
                # ('datetimeevents', 'hadm_id', (1, 2)),
                ("diagnoses_icd", "hadm_id", (1, 2)),
                ("drgcodes", "hadm_id", (1, 2)),
                ("icustays", "hadm_id", (1, 2)),
                ("procedures_icd", "hadm_id", (1, 2)),
                ("services", "hadm_id", (1, 2)),
                ("documents", "hadm_id", (1, 2)),
                ("transfers", "hadm_id", (1, 2)),
                ("prescriptions", "hadm_id", (1, 2)),
                ("labevents", "hadm_id", (1, 2)),
                ("microbiologyevents", "hadm_id", (1, 2)),
            ],
        },
        "subject_id": {
            "parents": [("patients", "subject_id"), ],
            "children": [
                ("admissions", "subject_id", (1, 2)),
                ("chartevents", "subject_id", (1, 2)),
                # ('datetimeevents', 'subject_id', (1, 2)),
                ("diagnoses_icd", "subject_id", (1, 2)),
                ("drgcodes", "subject_id", (1, 2)),
                ("icustays", "subject_id", (1, 2)),
                ("procedures_icd", "subject_id", (1, 2)),
                ("services", "subject_id", (1, 2)),
                ("transfers", "subject_id", (1, 2)),
                ("prescriptions", "subject_id", (1, 2)),
                ("labevents", "subject_id", (1, 2)),
                ("microbiologyevents", "subject_id", (1, 2)),
                ("documents", "patient_id", (1, 2)),
            ],
        },
        "cgid": {
            "parents": [("caregivers", "cgid"), ],
            "children": [
                ("chartevents", "cgid", (1, 2)),
                # ('datetimeevents', 'cgid', (1, 2)),
                ("documents", "cgid", (1, 2)),
            ],
        },
        "icustay_id": {
            "parents": [("icustays", "icustay_id"), ],
            "children": [
                ("chartevents", "icustay_id", (0, 2)),
                # ('datetimeevents', 'icustay_id', (1, 2)),
                ("transfers", "icustay_id", (1, 2)),
                ("prescriptions", "icustay_id", (1, 2)),
            ],
        },
        "itemid": {
            "parents": [("d_items", "itemid"), ],
            "children": [
                ("chartevents", "itemid", (1, 2)),
                # ('datetimeevents', 'itemid', (1, 2)),
                ("labevents", "itemid", (1, 2)),
                ("microbiologyevents", "spec_itemid", (1, 2)),
                ("microbiologyevents", "org_itemid", (1, 2)),
                ("microbiologyevents", "ab_itemid", (1, 2)),
            ],
        },
        "icd9_code_diagnoses": {
            "parents": [("d_icd_diagnoses", "icd9_code"), ],
            "children": [("diagnoses_icd", "icd9_code", (1, 2)), ],
        },
        "icd9_code_procedures": {
            "parents": [("d_icd_procedures", "icd9_code"), ],
            "children": [("procedures_icd", "icd9_code", (1, 2)), ],
        },
        "service": {
            "parents": [("hospit_services", "careunit")],
            "children": [
                ("icustays", "first_careunit", (1, 2)),
                ("icustays", "last_careunit", (1, 2)),
                ("services", "prev_service", (1, 2)),
                ("services", "curr_service", (1, 2)),
                ("transfers", "prev_careunit", (1, 2)),
                ("transfers", "curr_careunit", (1, 2)),
            ],
        },
    }

    db_gen = DataBaseRandomGenerator(locale=args.locale, translation_enabled=args.translate)

    for table_name, table in tables.items():
        db_gen.add_table(table_name, table)

    for relation_name, relation_dict in relations.items():
        db_gen.add_relation(relation_name, relation_dict)

    db_gen.random_generation(min_rows=args.rows, max_card=5)

    for table_name, table in tqdm(tables.items(), desc="Uploading to postgres"):
        db_gen.tables[table_name].to_sql(table_name, engine)


def generate_pdf():
    parser = argparse.ArgumentParser(
        description="Generates prescriptions PDF based on a postgres MIMIC database"
    )
    parser.add_argument(
        "--db",
        required=True,
        type=str,
        help="the database connexion string (eg: postgresql://user:passsword@host/database)",
    )
    parser.add_argument(
        "--locale",
        default="fr_FR",
        type=str,
        help="The locale to use when generating data",
        # see available locales here (https://faker.readthedocs.io/en/master/locales.html)
    )

    args = parser.parse_args()

    print("Connecting to postgres...")
    engine = create_engine(args.db)
    connection = engine.connect()
    print("OK")

    result = connection.execute(
        """SELECT
            caregivers.first_name AS doctor_first_name,
            caregivers.last_name AS doctor_last_name,
            patients.first_name AS patient_first_name,
            patients.last_name AS patient_last_name,
            patients.dob AS patient_birth_date,
            patients.weight AS patient_weight,
            d_icd_diagnoses.short_title AS diagnoses_st,
            d_icd_diagnoses.long_title AS diagnoses_lt,
            documents.document as path,
            documents.type as type,
            documents.hadm_id as hadm_id
        FROM documents
        JOIN caregivers ON documents.cgid = caregivers.cgid
        JOIN patients ON documents.patient_id = patients.subject_id
        JOIN diagnoses_icd ON documents.patient_id = diagnoses_icd.subject_id JOIN d_icd_diagnoses ON diagnoses_icd.icd9_code = d_icd_diagnoses.icd9_code"""
    )

    for row in tqdm(result, desc="Generating PDF prescriptions from database"):
        doc_date_result = connection.execute(
            f"""SELECT startdate as doc_date FROM prescriptions WHERE hadm_id LIKE '{row['hadm_id']}'"""
        )
        for date in doc_date_result:
            doc_date = datetime.datetime.strptime(date["doc_date"], "%Y-%m-%d %H:%M:%S")
            patient_birth_date = datetime.datetime.strptime(
                row["patient_birth_date"], "%Y-%m-%d %H:%M:%S"
            )
            patient_age = (
                    doc_date.year
                    - patient_birth_date.year
                    - (
                            (doc_date.month, doc_date.day)
                            < (patient_birth_date.month, patient_birth_date.day)
                    )
            )
            medication_result = connection.execute(
                f"""SELECT
                        drug as name,
                        prod_strength as dose,
                        dose_val_rx as poso_qty,
                        dose_unit_rx as poso_type
                    FROM
                        prescriptions
                    WHERE
                        hadm_id LIKE '{row['hadm_id']}' AND startdate LIKE '{doc_date}'
                """
            )
            medication = []
            for medic in medication_result:
                medication.append(
                    {
                        "name": medic["name"],
                        "dose": medic["dose"],
                        "poso_qty": medic["poso_qty"],
                        "poso_type": medic["poso_type"],
                    }
                )
        if row["type"] == "prescription":
            prescription_generator(
                row["path"],
                locale=args.locale,
                doc_attributes={"date": doc_date.strftime("%d/%m/%Y")},
                doctor={"first_name": row["doctor_first_name"], "last_name": row["doctor_last_name"], },
                patient={
                    "first_name": row["patient_first_name"],
                    "last_name": row["patient_last_name"],
                    "age": patient_age,
                    "weight": row["patient_weight"],
                },
                medication=medication,
            )
        elif row["type"] == "report":
            cr_generator(
                row["path"],
                locale=args.locale,
                doctor={"first_name": row["doctor_first_name"], "last_name": row["doctor_last_name"], },
                patient={
                    "first_name": row["patient_first_name"],
                    "last_name": row["patient_last_name"],
                    "age": patient_age,
                    "weight": row["patient_weight"],
                },
                hospitalization={
                    "reason": row["diagnoses_st"],
                    "hospitalization": row["diagnoses_lt"]
                },
                date={
                    "last_date": doc_date.strftime("%d/%m/%Y"),
                    "first_date": (doc_date - datetime.timedelta(days=randint(0, 16))).strftime("%d/%m/%Y"),
                    "birth_date": patient_birth_date
                },
                medication=medication,
            )


import os
from setuptools import setup, find_packages

with open(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md"), encoding="utf-8",
) as f:
    long_description = f.read()


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


requirements = read("requirements.txt").split()

setup(
    name="pysin",
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "pysin-gen-db=pysin.mimic.cli:generate_database",
            "pysin-gen-pdf=pysin.mimic.cli:generate_pdf",
            "pysin-preprocess=pysin.mimic.cli:preprocess",
            "pysin-scrap=pysin.mimic.cli:scrap",
        ]
    },
    version="1.5.1",
    license="Apache License 2.0",
    description="PySin is a toolbox for text retrieval in unstructured documents datasets. It contains both a multi-type text extractor and a search engine. To test them, you can use the medical prescriptions generator that is also provided.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Jean-Baptiste Laval",
    author_email="contact@arkhn.com",
    url="https://github.com/arkhn/PySin",
    download_url="https://github.com/arkhn/PySin/archive/1.1.0.tar.gz",
    keywords=[
        "arkhn",
        "text retrieval",
        "search engine",
        "text extraction",
        "dataset generator",
        "medical",
    ],
    install_requires=requirements,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)



import os
import sys
import csv
import re
import argparse
import json
import shutil
import subprocess

from path import Path

import pdftotext
import docx2txt
from striprtf.striprtf import rtf_to_text


# The types that can be handled are : csv, doc, docx, html, md, pdf, rtf, txt, xml


def extract(path):
    """Extract texts from documents in the given folder and its subfolders and store them at the same location in files called '.<TYPE>-----<NAME>.txt'."

    Arguments:
        path {str} -- path to the folder where the files to extract are

    Returns:
        None
    """

    path = Path(path)

    extract_folders(path, first_call=True)


def extract_folders(path, first_call=False):
    """Recursively extracts text

    Arguments:
        path {Path} -- path to the folder where the files to extract are

    Keyword Arguments:
        first_call {bool} -- for monitoring purpose : must be True if the current call is not recursive (default: {False})

    Returns:
        None
    """

    if first_call:
        print("New extraction")
        count_folder = 0

    targets = os.listdir(path)

    for f in targets:

        current_path = path / f

        if os.path.isdir(current_path):

            if first_call:
                print(f"Loading : {int(100 * count_folder / len(targets))} %", end="\r")
                count_folder += 1

            extract_folders(current_path)

        elif not re.match(r"\.\w{2,4}\-{5}.*", f):

            name, file_type = re.findall(r"^(.+)\.(\w{2,4})$", f.lower())[0]

            if not os.path.isfile(path / f".{file_type}-----{name}.txt"):

                if file_type == "pdf":

                    try:

                        extract_pdf2txt(
                            current_path, path / f".pdf-----{name}.txt",
                        )

                    except (pdftotext.Error):
                        txt_file_path = path / f".pdf-----{name}.txt"
                        if os.path.exists(txt_file_path):
                            os.remove(txt_file_path)

                elif file_type == "docx":

                    extract_docx2txt(
                        current_path, path / f".docx-----{name}.txt",
                    )

                elif file_type == "doc":

                    extract_doc2txt(
                        current_path, path / f".doc-----{name}.txt",
                    )

                elif file_type == "rtf":

                    extract_rtf2txt(
                        current_path, path / f".rtf-----{name}.txt",
                    )

                elif file_type in {"csv", "html", "md", "txt", "xml"}:

                    shutil.copyfile(current_path, path / f".{file_type}-----{name}.txt")

    if first_call:
        print("Extraction terminated !")


def extract_pdf2txt(input_path, output_path):
    """Extracts the text from a pdf document and store it in a txt document

    Arguments:
        input_path {Path} -- pdf file to extract text from
        output_path {Path} -- txt file to store text in

    Returns:
        bool -- True if the pdf document could not be processed (probably because it's a scanned document)
    """

    with open(input_path, "rb") as f:
        pdf = pdftotext.PDF(f)
    if all(page == "" for page in pdf):
        return True
    with open(output_path, "w+") as f:
        for page in pdf:
            f.write(page)
            f.write("\n\n$$END OF PAGE$$\n\n")
    return False


def extract_docx2txt(input_path, output_path):
    """Extracts the text from a docx document and store it in a txt document

    Arguments:
        input_path {Path} -- docx file to extract text from
        output_path {Path} -- txt file to store text in

    Returns:
        None
    """

    with open(output_path, "w+") as f:
        f.write(docx2txt.process(input_path))


def extract_doc2txt(input_path, output_path):
    """Extracts the text from a doc document and store it in a txt document. It creates a tempoary docx file from the doc one and removes it after extraction.

    Arguments:
        input_path {Path} -- doc file to extract text from
        output_path {Path} -- txt file to store text in

    Returns:
        None
    """

    subprocess.call(
        [
            "sudo",
            "lowriter",
            "--headless",
            "--convert-to",
            "docx",
            "--outdir",
            f"{input_path.rsplit('/', 1)[0]}",
            f"{input_path}",
        ]
    )
    extract_docx2txt(f"{input_path}x", output_path)
    os.remove(f"{input_path}x")


def extract_rtf2txt(input_path, output_path):
    """Extracts the text from a rtf document and store it in a txt document

    Arguments:
        input_path {Path} -- rtf file to extract text from
        output_path {Path} -- txt file to store text in

    Returns:
        None
    """

    with open(output_path, "w+",) as f:
        f.write(rtf_to_text(open(input_path).read()))


def reset_cache(path):
    """Clear all cached files in a folder. The files to erase are called '.<TYPE>-----<NAME>.txt'.

    Arguments:
        path {Path} -- folder to clear the cached files in

    Returns:
        None
    """

    path = Path(path)

    for f in os.listdir(path):

        current_path = path / f

        if os.path.isdir(current_path):

            reset_cache(current_path)

        elif re.match(r"\.\w{2,4}\-{5}.*", f):

            os.remove(current_path)


if __name__ == "__main__":

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "path", metavar="P", type=str, help="the path to the folder to preprocess"
    )
    parser.add_argument(
        "--reset", dest="reset", action="store_true", help="to clean the cached files"
    )

    args = parser.parse_args()

    reset_cache(args.path) if args.reset else extract(args.path)



from faker import Factory
from random import choice, choices, randint, random, sample, shuffle
from tqdm import tqdm
from google.cloud import translate_v2 as translate


class DataBaseRandomGenerator:
    def __init__(self, locale="fr_FR", translation_enabled=True):
        self._fake_factory = Factory.create(locale)
        self.locale = locale
        self.tables = {}
        self._possible_values = {}
        self._possible_length = {}
        self.relations = {}
        self._translator = translate.Client() if translation_enabled else None

    def add_table(self, name, table):
        assert not name in self.tables, f"The table {name} already exists."
        self.tables[name] = table
        self._possible_values[name] = {}
        self._possible_length[name] = 0

    def add_relation(self, name, relation_dict):
        assert (
            type(relation_dict) == dict
            and "parents" in relation_dict.keys()
            and "children" in relation_dict.keys()
        ), "A relation must be a dict following the pattern {'parents': (table, attribute), 'children': (table, attribute, type)}."
        test_table_name, test_attr_name = relation_dict["parents"][0]
        attribute_type = self.tables[test_table_name].attributes[test_attr_name].attr_type.type_name
        assert name not in self.relations, f"The relation {name} already exists."
        assert (
            test_table_name in self.tables.keys()
        ), f"A relation uses the table {test_table_name} whereas this table does not exist."
        for table_name, attr_name in relation_dict["parents"]:
            assert (
                table_name in self.tables.keys()
            ), f"A relation uses the table {table_name} whereas this table does not exist."
            assert (
                attr_name in self.tables[table_name].attributes.keys()
            ), f"A relation uses the attribute {attr_name} from table {table_name} whereas this attribute does not exist."
            assert (
                self.tables[table_name].attributes[attr_name].attr_type.type_name == attribute_type
            ), f"The following attributes are in relation but their types are different : {attr_name} from {table_name} (of type {self.tables[table_name].attributes[attr_name].attr_type.type_name}) and {test_attr_name} from {test_table_name} (of type {attribute_type})."
        for table_name, attr_name, _ in relation_dict["children"]:
            assert (
                table_name in self.tables.keys()
            ), f"A relation uses the table {table_name} whereas this table does not exist."
            assert (
                attr_name in self.tables[table_name].attributes.keys()
            ), f"A relation uses the attribute {attr_name} from table {table_name} whereas this attribute does not exist."
            assert (
                self.tables[table_name].attributes[attr_name].attr_type.type_name == attribute_type
            ), f"The following attributes are in relation but their types are different : {attr_name} from {table_name} (of type {self.tables[table_name].attributes[attr_name].attr_type.type_name}) and {test_attr_name} from {test_table_name} (of type {attribute_type})."
        self.relations[name] = relation_dict

    def _satisfy_relation(self, relation_dict, min_rows=10, max_card=5, missing_rate=0.1):
        test_table_name, test_attr_name = relation_dict["parents"][0]
        length = max(
            [self._possible_length[table_name] for table_name, _ in relation_dict["parents"]]
        )
        attribute_class = self.tables[test_table_name].attributes[test_attr_name]
        for table_name, attr_name in relation_dict["parents"]:
            assert self._possible_length[table_name] in {
                0,
                length,
            }, f"The tables {test_table_name} and {table_name} are in relation one-to-one with the attributes {test_attr_name} = {attr_name} but their length are different : {length} != {self._possible_length[table_name]}."
        if not length:
            length = min_rows
        possible_values = []
        if attribute_class.attr_type.type_name == "field":
            try:
                possible_values = sample(
                    attribute_class.kwargs_rd_gen["possible_values"].to_list(), k=length
                )
            except ValueError:
                print(
                    f"[{test_table_name}.{test_attr_name}] The value of 'min_rows' is too big regarding the size of the input database."
                    f" Downsizing to {len(attribute_class.kwargs_rd_gen['possible_values'])} rows"
                )
                possible_values = sample(
                    attribute_class.kwargs_rd_gen["possible_values"].to_list(),
                    k=len(attribute_class.kwargs_rd_gen["possible_values"]),
                )
        else:
            for _ in range(length):
                val = attribute_class.get_random_value(
                    self._fake_factory, self._translator, self.locale
                )
                while val in possible_values:
                    val = attribute_class.get_random_value(
                        self._fake_factory, self._translator, self.locale
                    )
                possible_values.append(val)
        length_possible_values = len(possible_values)
        for table_name, attr_name in relation_dict["parents"]:
            self._possible_values[table_name][attr_name] = possible_values
            self._possible_length[table_name] = length_possible_values
        for table_name, attr_name, rel_type in relation_dict["children"]:
            length = self._possible_length[table_name]
            assert rel_type != (2, 2), "Cannot handle (2,2) relations yet."
            if rel_type == (0, 1):
                if length == 0:
                    self._possible_values[table_name][attr_name] = []
                    nb_values = 0
                    for val in possible_values:
                        if random() > missing_rate:
                            self._possible_values[table_name][attr_name].append(val)
                            nb_values += 1
                    self._possible_length[table_name] = nb_values
                else:
                    assert (
                        length <= length_possible_values
                    ), f"Not enough values to fit in attribute {attr_name} from {table_name}."
                    self._possible_values[table_name][attr_name] = choices(
                        possible_values, k=length
                    )
            elif rel_type == (0, 2):
                shuffle(possible_values)
                current_possible_values = possible_values[
                    : int(length_possible_values * (1 - missing_rate))
                ]
                if length == 0:
                    self._possible_values[table_name][attr_name] = choices(
                        current_possible_values, k=int(length_possible_values * max_card / 2),
                    )
                    self._possible_length[table_name] = len(current_possible_values)
                else:
                    self._possible_values[table_name][attr_name] = choices(
                        current_possible_values, k=length
                    )
            elif rel_type == (1, 2):
                if length == 0:
                    current_possible_values = possible_values + choices(
                        possible_values, k=int(length_possible_values * (max_card / 2 - 1)),
                    )
                    shuffle(current_possible_values)
                    self._possible_values[table_name][attr_name] = current_possible_values
                    self._possible_length[table_name] = len(current_possible_values)
                else:
                    assert (
                        length >= length_possible_values
                    ), f"Too many values to fit in attribute {attr_name} from table {table_name}."
                    current_possible_values = possible_values + choices(
                        possible_values, k=length - length_possible_values
                    )
                    shuffle(current_possible_values)
                    self._possible_values[table_name][attr_name] = current_possible_values

    def random_generation(self, min_rows=100, max_card=5, missing_rate=0.1):

        print("Starts filling table...")
        nb_relations = len(self.relations)
        r = 0
        for _, relation_dict in tqdm(self.relations.items(), desc="Satisfying relations"):
            self._satisfy_relation(
                relation_dict, min_rows=min_rows, max_card=max_card, missing_rate=missing_rate,
            )
            r += 1
        nb_tables = len(self.tables)
        t = 0
        for table in tqdm(self.tables, desc="Filling tables"):
            if self._possible_length[table]:
                table_length = self._possible_length[table]
                for _, possible_val in self._possible_values[table].items():
                    assert (
                        len(possible_val) == table_length
                    ), f"Two columns are supposed to be filled with values lists of different sizes in table {table}. --> {len(possible_val)} and {table_length}"
            else:
                table_length = min_rows
                self._possible_length[table] = min_rows
            for i in range(table_length):
                self.tables[table].add_random_row(
                    self._fake_factory,
                    self._translator,
                    self.locale,
                    fixed_values={
                        attr: values[i] for attr, values in self._possible_values[table].items()
                    },
                )
            t += 1
        print("\nDatabase filled !")
